/*
Szymon Rusinkiewicz
Princeton University

colormap.cc
Define colormaps for the xshade application
*/

#include "GL/glui.h"
#include "Vec.h"
#include "colormap.h"
#include <vector>
using std::vector;

#ifndef GL_CLAMP_TO_EDGE
#define GL_CLAMP_TO_EDGE 0x812F
#endif

// Global vars
int colormap = COLORMAP_BW;
float Gamma = 1.6f;


// Set up the glClearColor to be used
void set_bgcolor()
{
	vec col = vec(0.9, 0.9, 0.9);
	if (colormap == COLORMAP_HIKEY)
		col = vec(1, 1, 1);
	else if (colormap == COLORMAP_SEPIA)
		col = vec(0.9, 0.89, 0.85);
	else if (colormap == COLORMAP_TAN)
		col = vec(1, 1, 1);
	else if (colormap == COLORMAP_GOOCHBASE)
		col = vec(0.54, 0.52, 0.50);

	for (int i = 0; i < 3; i++)
		col[i] = pow(col[i], 1.0f / Gamma);
	glClearColor(col[0], col[1], col[2], 1.0f);
}

vector< vector<float> > textures;

// Set up texture map for the rendering.  These are 2-D textures that in
// most cases only vary along 1 dimension, representing "darkest" to
// "lightest" shading.  A few colormaps have the full 2-D worth of
// indexing, computed with lightsources 90 degrees apart.
void setuptexture()
{
	// Texture map size - resources/quality tradeoff
	const int texsize = 512;

	// If first time, allocate texture contexts
	static GLuint *texcontexts = NULL;
	static float *computed_at_Gamma = NULL;
	if (!texcontexts) {
		texcontexts = new GLuint[ncolormaps];
		glGenTextures(ncolormaps, texcontexts);
		computed_at_Gamma = new float[ncolormaps];
		for (int i = 0; i < ncolormaps; i++)
			computed_at_Gamma[i] = 0;
		textures.resize(ncolormaps);
	}

	// Remap texture coordinates from [-1..1] to [0..1]
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glTranslatef(0.5, 0.5, 0.5);
	glScalef(0.5, 0.5, 0.5);
	glMatrixMode(GL_MODELVIEW);

	// Bind the texture, and set up misc parameters
	glBindTexture(GL_TEXTURE_2D, texcontexts[colormap]);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_TEXTURE_2D);

	// If the current texture is valid, that's all we have to do
	if (computed_at_Gamma[colormap] == Gamma)
		return;


	// OK, we'll have to build this texture
	int tex_x = texsize, tex_y = 1;
	if (colormap >= COLORMAP_RAINBOW)
		tex_y = texsize;
	//vector<float> texture(3*tex_x*tex_y);
	vector<float> &texture = textures[colormap];
	texture.resize(3*tex_x*tex_y);

	// Specify points on the colormap curves: the colors are
	// interpolated with parabolas through (c1,c2,c3) and (c3,c4,c5)
	vec c1x = vec(1, 1, 1);
	vec c2x = vec(0.75, 0.75, 0.75);
	vec c3x = vec(0.5, 0.5, 0.5);
	vec c4x = vec(0.25, 0.25, 0.25);
	vec c5x = vec(0, 0, 0);

	if (colormap == COLORMAP_HIKEY) {
		c1x = vec(1, 1, 1);
		c3x = vec(0.75, 0.75, 0.75);
		c5x = vec(0.5, 0.5, 0.5);
		c2x = 0.25f * c3x + 0.75f * c1x;
		c4x = 0.25f * c3x + 0.75f * c5x;
	} else if (colormap == COLORMAP_GOOCH) {
		c1x = vec(1, 1, 0.7);
		c2x = vec(0.75, 0.7, 0.6);
		c3x = vec(0.5, 0.5, 0.5);
		c4x = vec(0.25, 0.22, 0.4);
		c5x = vec(0, 0, 0.3);
	} else if (colormap == COLORMAP_SEPIA) {
		//c1x = vec(1, 1, 1);
		c1x = vec(0.9, 0.89, 0.85);
		c2x = vec(0.75, 0.7, 0.62);
		c3x = c2x * c2x;
		c4x = c3x * c3x;
		c5x = c4x * c4x;
	} else if (colormap == COLORMAP_TAN) {
		c1x = vec(1, 0.97, 0.94);
		c3x = vec(0.65, 0.60, 0.52);
		c5x = vec(0.1, 0.08, 0.13);
		c2x = 0.7f * c3x + 0.4f * c1x;
		c4x = 0.5f * c3x + 0.4f * c5x;
	} else if (colormap == COLORMAP_RAINBOW) {
		c1x = vec(1, 0.93, 0.5);
		c3x = vec(0.5, 0.5, 0.5);
		c5x = vec(0, 0, 0.4);
		c2x = 0.5f * c3x + 0.5f * c1x;
		c4x = 0.5f * c3x + 0.5f * c5x;
	} else if (colormap == COLORMAP_GREENBROWN) {
		c1x = vec(0.6, 0.6, 0.6);
		c3x = vec(0, 0, 0);
		c5x = vec(-0.9, -0.9, -0.9);
		c2x = 0.67f * c3x + 0.33f * c1x;
		c4x = 0.67f * c3x + 0.33f * c5x;
	} else if (colormap == COLORMAP_GOOCHBASE) {
		c1x = vec(1.26, 1.15, 1.0);
		c3x = vec(0.77, 0.76, 0.76);
		c5x = vec(0.36, 0.35, 0.46);
		c2x = 0.25f * c3x + 0.75f * c1x;
		c4x = 0.25f * c3x + 0.75f * c5x;
	}

	vec c1y = vec(1, 1, 1);
	vec c2y = vec(0.75, 0.75, 0.75);
	vec c3y = vec(0.5, 0.5, 0.5);
	vec c4y = vec(0.25, 0.25, 0.25);
	vec c5y = vec(0, 0, 0);
	if (colormap == COLORMAP_RAINBOW) {
		c1y = vec(1.5, 1.0, 1.0);
		c5y = vec(0.5, 1.0, 1.0);
		c3y = 0.5f * c1y + 0.5f * c5y;
		c2y = 0.5f * c3y + 0.5f * c1y;
		c4y = 0.5f * c3y + 0.5f * c5y;
	} else if (colormap == COLORMAP_GREENBROWN) {
		c1y = vec(0.2, 0.45, 0.3);
		c3y = vec(0.4, 0.4, 0.3);
		c5y = vec(0.2, 0.1, 0.0);
		c2y = 0.5f * c3y + 0.5f * c1y;
		c4y = 0.5f * c3y + 0.5f * c5y;
	} else if (colormap == COLORMAP_GOOCHBASE) {
		c1y = vec(0.9, 0.89, 0.85);
		c2y = vec(0.75, 0.7, 0.62);
		c3y = c2y * c2y;
		c4y = c3y * c3y;
		c5y = c4y * c4y;
	}

	// Compute parameters of the parabolas
	vec a1x = 2.0f * c1x - 4.0f * c2x + 2.0f * c3x;
	vec b1x =       -c1x + 4.0f * c2x - 3.0f * c3x;
	vec a2x = 2.0f * c5x - 4.0f * c4x + 2.0f * c3x;
	vec b2x =       -c5x + 4.0f * c4x - 3.0f * c3x;
	vec a1y = 2.0f * c1y - 4.0f * c2y + 2.0f * c3y;
	vec b1y =       -c1y + 4.0f * c2y - 3.0f * c3y;
	vec a2y = 2.0f * c5y - 4.0f * c4y + 2.0f * c3y;
	vec b2y =       -c5y + 4.0f * c4y - 3.0f * c3y;

	// Go through and build the textures
	float *t = &texture[0];
	for (int j = 0; j < tex_y; j++) {
		float y = (tex_y == 1) ? 0.0f : 2.0f * j / (tex_y - 1) - 1.0f;
		vec coly = c3y;
		if (y >= 0.0f) {
			if (y > 1.0f)
				y = 1.0f;
			coly += y*y * a1y;
			coly += y * b1y;
		} else {
			if (y < -1.0f)
				y = -1.0f;
			coly += y*y * a2y;
			coly -= y * b2y;
		}
		for (int i = 0; i < tex_x; i++) {
			float x = 2.0f * i / (tex_x - 1) - 1.0f;
			vec col = c3x;
			if (x >= 0.0f) {
				if (x > 1.0f)
					x = 1.0f;
				col += x*x * a1x;
				col += x * b1x;
			} else {
				if (x < -1.0f)
					x = -1.0f;
				col += x*x * a2x;
				col -= x * b2x;
			}
			if (colormap == COLORMAP_RAINBOW) {
				col *= coly;
			} else if (colormap == COLORMAP_GREENBROWN) {
				if (col[0] < 0.0f)
					col = (col + vec(1,1,1)) * coly;
				else
					col += (vec(1,1,1) - col) * coly;
			} else if (colormap == COLORMAP_GOOCHBASE) {
				col *= coly;
			}
			*t++ = col[0];
			*t++ = col[1];
			*t++ = col[2];
		}
	}

	// Apply gamma
	float one_over_Gamma = 1.0f / Gamma;
	for (int i = 0; i < 3*tex_x*tex_y; i++)
		texture[i] = pow(texture[i], one_over_Gamma);
	computed_at_Gamma[colormap] = Gamma;

	// Set the texture
	glTexImage2D(GL_TEXTURE_2D, 0, 3, tex_x, tex_y, 0,
		     GL_RGB, GL_FLOAT, &texture[0]);
}


// Create the parts of the GUI that contain the colormap stuff
void make_colormap_gui(GLUI *glui, GLUI_Panel *p,
		       bool include_gamma /* = true */)
{
	glui->add_statictext_to_panel(p, "Colormap:");
	GLUI_RadioGroup *r = glui->add_radiogroup_to_panel(p, &colormap);
	glui->add_radiobutton_to_group(r, "Gray");
	glui->add_radiobutton_to_group(r, "Low-contrast");
	glui->add_radiobutton_to_group(r, "Gooch (yellow-blue)");
	glui->add_radiobutton_to_group(r, "Sepia");
	glui->add_radiobutton_to_group(r, "Tan + gray");
	glui->add_radiobutton_to_group(r, "Rainbow");
	glui->add_radiobutton_to_group(r, "Green-brown");
	glui->add_radiobutton_to_group(r, "Gooch for base coat");

	if (include_gamma)
		glui->add_slider_to_panel(p, "Gamma", GLUI_SLIDER_FLOAT,
					  0.3, 3.0, &Gamma);
}

