/*
Szymon Rusinkiewicz
Princeton University

shade.cc
Exaggerated shading computation.  This is the CPU-based version.
*/

#include "shade.h"
using namespace std;

// Compute an s-shaped curve, clamping to [-thresh..thresh]
static inline float s(float a, float thresh)
{
	// Option 1: Piecewise linear
	return (a < -thresh) ? -thresh : (a > thresh) ? thresh : a;

	// Option 2: Piecewise quadratic
	//float aa = fabs(a);
	//if (aa >= 2.0f * thresh)
	//	return (a >= 0.0f) ? thresh : -thresh;
	//return a - 0.25f * a * aa / thresh;
}


// Simple cosine shading.  Also used for the base coat...
void shade_cosine(const vector<vec> &normals, const vec &lightdir,
		  float mult, float thresh,
		  vector<float> &texcoords, int secondpass)
{
	vec ml = mult * lightdir;

	int nv = normals.size();
#pragma omp parallel for
	for (int i = 0; i < nv; i++)
		texcoords[2*i+secondpass] = s(ml DOT normals[i], thresh);
}


// Contour shading - brings out occluding contours
void shade_cntr(const vector<vec> &normals, const vec &viewdir, float strength,
		  vector<float> &texcoords, int secondpass)
{
	float mult = 1.0f / exp(-strength);
	int nv = normals.size();
	for (int i = 0; i < nv; i++) {
		float old_t = 0.5f + 0.5f * texcoords[2*i+secondpass];
		float ndotv = max(normals[i] DOT viewdir, 1.0e-10f);
		float t = mult * exp(-strength / ndotv);
		texcoords[2*i+secondpass] = 2.0f * old_t * t - 1.0f;
	}
}


// Exaggerated shading
void shade_exaggerated(const TriMesh *mesh, const XSHdata *data,
		       const vec &lightdir, int use_pdir,
		       const vector<float> &mult, const vector<float> &thresh,
		       const point &attention, float attention_r,
		       vector<float> &texcoords, int secondpass)
{
	int nv = mesh->vertices.size();
	int nn = data->sigmas.size() - 1;
#pragma omp parallel for schedule(dynamic,1)
	for (int pass = 0; pass < nn; pass++) {
		const vector<vec> &n = data->snormals[pass];
		const vector<vec> &nf = data->snormals[pass+1];
		const vector<vec> &p = data->spdir[pass+1];
		float currmult = mult[pass];
		float currthresh = thresh[pass];
		float atten_mult = 5.0f * data->sigmas[0] / data->sigmas[pass] /
			sqr(attention_r);
		if (!use_pdir) {
			for (int i = 0; i < nv; i++) {
				float nfl = nf[i] DOT lightdir;
				float t = (n[i] DOT lightdir) -
					nfl * (n[i] DOT nf[i]);
				t /= sqrt(1.0f - nfl*nfl);
				t = s(t * currmult, currthresh);
				if (attention_r != 0.0f)
					t /= 1.0f + atten_mult *
					     dist2(mesh->vertices[i], attention);
				texcoords[2*i+secondpass] += t;
			}
		} else {
			const float pdir_mult = 10.0f;
			for (int i = 0; i < nv; i++) {
				float nfl = nf[i] DOT lightdir;
				float pl = p[i] DOT lightdir;
				pl *= pdir_mult;
				vec lp(lightdir[0] - nfl*nf[i][0] + pl*p[i][0],
				       lightdir[1] - nfl*nf[i][1] + pl*p[i][1],
				       lightdir[2] - nfl*nf[i][2] + pl*p[i][2]);
				float t = (n[i] DOT lp) / len(lp);
				t = s(t * currmult, currthresh);
				if (attention_r != 0.0f)
					t /= 1.0f + atten_mult *
					     dist2(mesh->vertices[i], attention);
				texcoords[2*i+secondpass] += t;
			}
		}
	}
}


// Multiscale mean-curvature shading
void shade_meancurv(const TriMesh *mesh, const XSHdata *data,
		    const vector<float> &mult, const vector<float> &thresh,
		    float curv_limit,
		    const point &attention, float attention_r,
		    vector<float> &texcoords, int secondpass)
{
	int nv = mesh->vertices.size();
	int nn = data->sigmas.size();
#pragma omp parallel for schedule(dynamic,1)
	for (int pass = 0; pass < nn; pass++) {
		const vector<float> &k = data->smeancurv[pass];
		float currmult = mult[pass];
		float currthresh = thresh[pass];
		float lthresh = currthresh * (-1.0f + curv_limit);
		float uthresh = currthresh * (1.0f + curv_limit);
		float atten_mult = 5.0f * data->sigmas[0] / data->sigmas[pass] /
			sqr(attention_r);
		for (int i = 0; i < nv; i++) {
			float t = min(max(k[i] * currmult, lthresh), uthresh);
			if (attention_r != 0.0f)
				t /= 1.0f + atten_mult *
				     dist2(mesh->vertices[i], attention);
			texcoords[2*i+secondpass] += t;
		}
	}
}

