/*
Szymon Rusinkiewicz
Princeton University

xsh.cc
The .xsh file format, which contains a mesh plus several levels of
pre-smoothed normals and curvatures stored in an efficient format.
*/

#include <stdio.h>
#include "xsh.h"
using namespace std;


// Pack normals into 32 bits
static void compress_normals(const vector<vec> &norm, vector<unsigned> &cnorm)
{
	const float add = (1u << 13u);
	const float mult = add * M_SQRT2;
	const unsigned m = (1u << 14u) - 1u;

	int nv = norm.size();
#pragma omp parallel for
	for (int i = 0; i < nv; i++) {
		float u = norm[i][1], v = norm[i][2];
		unsigned b = (norm[i][0] < 0.0f);
		if (fabs(norm[i][1]) > fabs(norm[i][2])) {
			if (fabs(norm[i][1]) > fabs(norm[i][0])) {
				u = norm[i][0];
				b = 2u | (norm[i][1] < 0.0f);
			}
		} else {
			if (fabs(norm[i][2]) > fabs(norm[i][0])) {
				v = norm[i][0];
				b = 4u | (norm[i][2] < 0.0f);
			}
		}
		unsigned U = min((unsigned)max(int(u * mult + add), 0), m);
		unsigned V = min((unsigned)max(int(v * mult + add), 0), m);

		cnorm[i] = (b << 28u) | (U << 14u) | V;
	}
}


// Unpack normals
static void uncompress_normals(vector<vec> &norm, const vector<unsigned> &cnorm)
{
	const float mult = M_SQRT2 / float(1u << 14u);
	const float add = mult * (0.5f - float(1u << 13u));
	const unsigned m = (1u << 14u) - 1u;
	const unsigned sm = (1u << 28u);

	int nv = cnorm.size();
	norm.resize(nv);
#pragma omp parallel for
	for (int i = 0; i < nv; i++) {
		float u = float((cnorm[i] >> 14u) & m) * mult + add;
		float v = float(cnorm[i] & m) * mult + add;
		float w = sqrt(1.0f - u*u - v*v);
		unsigned b = (cnorm[i] >> 29u);
		if (b == 0u) {
			norm[i][0] = (cnorm[i] & sm) ? -w : w;
			norm[i][1] = u;
			norm[i][2] = v;
		} else if (b == 1u) {
			norm[i][0] = u;
			norm[i][1] = (cnorm[i] & sm) ? -w : w;
			norm[i][2] = v;
		} else {
			norm[i][0] = v;
			norm[i][1] = u;
			norm[i][2] = (cnorm[i] & sm) ? -w : w;
		}
		normalize(norm[i]);
	}
}


// Pack pdir into 16 bits
static void compress_pdir(const vector<vec> &pdir,
			  vector<unsigned short> &cpdir,
			  const vector<vec> &norm,
			  const vector<unsigned> &cnorm)
{
	int nv = pdir.size();
#pragma omp parallel for
	for (int i = 0; i < nv; i++) {
		vec t(1, 0, 0);
		if ((cnorm[i] >> 29u) == 0u) {
			t = vec(0, 1, 0);
		} else if ((cnorm[i] >> 29u) == 1u) {
			t = vec(0, 0, 1);
		}
		t -= norm[i] * (t DOT norm[i]);
		normalize(t);
		vec b = norm[i] CROSS t;

		float u = pdir[i] DOT t;
		float v = pdir[i] DOT b;
		unsigned U = min(max(int(u * 127.0f + 127.5f), 0), 254);
		unsigned V = min(max(int(v * 127.0f + 127.5f), 0), 254);

		cpdir[i] = (U << 8u) | V;
	}
}


// Unpack pdir
static void uncompress_pdir(vector<vec> &pdir,
			    const vector<unsigned short> &cpdir,
			    const vector<vec> &norm,
			    const vector<unsigned> &cnorm)
{
	const float mult = 1.0f / 127.0f;

	int nv = cpdir.size();
	pdir.resize(nv);
#pragma omp parallel for
	for (int i = 0; i < nv; i++) {
		vec t(1, 0, 0);
		if ((cnorm[i] >> 29u) == 0u) {
			t = vec(0, 1, 0);
		} else if ((cnorm[i] >> 29u) == 1u) {
			t = vec(0, 0, 1);
		}
		t -= norm[i] * (t DOT norm[i]);
		normalize(t);
		vec b = norm[i] CROSS t;

		float u = float(cpdir[i] >> 8u) * mult - 1.0f;
		float v = float(cpdir[i] & 255u) * mult - 1.0f;
		pdir[i] = u * t + v * b;
	}
}


// Pack H into 16 bits
static void compress_meancurv(const vector<float> &meancurv,
			      vector<unsigned short> &cmeancurv)
{
	int nv = meancurv.size();
#pragma omp parallel for
	for (int i = 0; i < nv; i++) {
		cmeancurv[i] = min(max(int(meancurv[i] * 1000.0f + 32767.5f),
					0), 65534);
	}
}


// Pack H into 16 bits
static void uncompress_meancurv(vector<float> &meancurv,
			        const vector<unsigned short> &cmeancurv)
{
	int nv = cmeancurv.size();
	meancurv.resize(nv);
#pragma omp parallel for
	for (int i = 0; i < nv; i++) {
		meancurv[i] = (float(cmeancurv[i]) - 32767.0f) * 0.001f;
	}
}


// Pack vertices into 3*24 bits each
static void compress_vertices(const vector<vec> &vertices,
			      vector<unsigned char> &cvertices,
			      const TriMesh::BSphere &bsphere)
{
	const float mult = float(1u << 24u) / (2.0f * bsphere.r);
	const float xadd = bsphere.r - bsphere.center[0];
	const float yadd = bsphere.r - bsphere.center[1];
	const float zadd = bsphere.r - bsphere.center[2];
	const unsigned m = (1u << 24u) - 1u;

	cvertices.clear();
	int nv = vertices.size();
	for (int i = 0; i < nv; i++) {
		unsigned X = min((unsigned)max(int((vertices[i][0]+xadd) * mult), 0), m);
		cvertices.push_back(X & 0xffu);
		X >>= 8u;
		cvertices.push_back(X & 0xffu);
		X >>= 8u;
		cvertices.push_back(X & 0xffu);
		unsigned Y = min((unsigned)max(int((vertices[i][1]+yadd) * mult), 0), m);
		cvertices.push_back(Y & 0xffu);
		Y >>= 8u;
		cvertices.push_back(Y & 0xffu);
		Y >>= 8u;
		cvertices.push_back(Y & 0xffu);
		unsigned Z = min((unsigned)max(int((vertices[i][2]+zadd) * mult), 0), m);
		cvertices.push_back(Z & 0xffu);
		Z >>= 8u;
		cvertices.push_back(Z & 0xffu);
		Z >>= 8u;
		cvertices.push_back(Z & 0xffu);
	}
}


// Unpack vertices
static void uncompress_vertices(vector<vec> &vertices,
			        const vector<unsigned char> &cvertices,
				const TriMesh::BSphere &bsphere)
{
	const float mult = (2.0f * bsphere.r) / float(1u << 24u);
	const float xadd = bsphere.center[0] - bsphere.r;
	const float yadd = bsphere.center[1] - bsphere.r;
	const float zadd = bsphere.center[2] - bsphere.r;
	int nv = cvertices.size();
	vertices.clear();
	vertices.reserve(nv);
	int i = 0;
	while (i < nv) {
		float x = cvertices[i] | (cvertices[i+1] << 8u) |
			(cvertices[i+2] << 16u);
		x = x * mult + xadd;
		i += 3;
		float y = cvertices[i] | (cvertices[i+1] << 8u) |
			(cvertices[i+2] << 16u);
		y = y * mult + yadd;
		i += 3;
		float z = cvertices[i] | (cvertices[i+1] << 8u) |
			(cvertices[i+2] << 16u);
		z = z * mult + zadd;
		i += 3;
		vertices.push_back(vec(x,y,z));
	}
}


// Pack tstrips
static void compress_tstrips(const vector<int> &tstrips,
			     vector<unsigned char> &ctstrips)
{
	const int maxoffs = 4;
	const int maxdist = 28;
	const int lit2 = 252;
	const int lit3 = 253;
	const int lit4 = 254;
	const int delim = 255;

	ctstrips.clear();
	int ns = tstrips.size();
	for (int i = 0; i < ns; i++) {
		if (tstrips[i] < 0) {
			ctstrips.push_back(delim);
			continue;
		}

		bool found = false;
		for (int j = 0; j < maxdist; j++) {
			int i2 = i - 1 - j;
			if (i2 < 0)
				break;
			if (tstrips[i2] >= 0 &&
			    abs(tstrips[i] - tstrips[i2]) <= maxoffs) {
				int offs = tstrips[i] - tstrips[i2];
				ctstrips.push_back(j + maxdist*(offs+maxoffs));
				found = true;
				break;
			}
		}
		if (found)
			continue;

		if (tstrips[i] <= 0xffff)
			ctstrips.push_back(lit2);
		else if (tstrips[i] <= 0xffffff)
			ctstrips.push_back(lit3);
		else
			ctstrips.push_back(lit4);

		ctstrips.push_back(tstrips[i] & 0xffu);
		ctstrips.push_back((tstrips[i] >> 8u) & 0xffu);
		if (tstrips[i] > 0xffffu)
			ctstrips.push_back((tstrips[i] >> 16u) & 0xffu);
		if (tstrips[i] > 0xffffffu)
			ctstrips.push_back(tstrips[i] >> 24u);
	}
}


// Unpack tstrips
static void uncompress_tstrips(vector<int> &tstrips,
			       const vector<unsigned char> &ctstrips)
{
	const int maxoffs = 4;
	const int maxdist = 28;
	const int lit2 = 252;
	const int lit3 = 253;
	const int lit4 = 254;
	const int delim = 255;

	int ns = ctstrips.size();
	tstrips.clear();
	tstrips.reserve(ns);
	for (int i = 0; i < ns; i++) {
		if (ctstrips[i] == delim) {
			tstrips.push_back(-1);
			continue;
		}
		if (ctstrips[i] == lit2) {
			unsigned s = ctstrips[i+1];
			s |= (ctstrips[i+2] << 8u);
			i += 2;
			tstrips.push_back(s);
			continue;
		}
		if (ctstrips[i] == lit3) {
			unsigned s = ctstrips[i+1];
			s |= (ctstrips[i+2] << 8u);
			s |= (ctstrips[i+3] << 16u);
			i += 3;
			tstrips.push_back(s);
			continue;
		}
		if (ctstrips[i] == lit4) {
			unsigned s = ctstrips[i+1];
			s |= (ctstrips[i+2] << 8u);
			s |= (ctstrips[i+3] << 16u);
			s |= (ctstrips[i+4] << 24u);
			i += 4;
			tstrips.push_back(s);
			continue;
		}

		int this_offs = (ctstrips[i] / maxdist) - maxoffs;
		int this_dist = ctstrips[i] % maxdist;
		int ind = tstrips.size() - 1 - this_dist;
		tstrips.push_back(this_offs + tstrips[ind]);
	}
}


// Byte swap uints, ints, and floats
static inline void swap_unsigned(unsigned &x)
{
	x = (x << 24) |
	    ((x << 8) & 0x00ff0000u) |
	    ((x >> 8) & 0x0000ff00u) |
	    (x >> 24);
}

static inline void swap_int(int &x)
{
	swap_unsigned(* (unsigned *)(&x));
}

static inline void swap_float(float &x)
{
	swap_unsigned(* (unsigned *)(&x));
}

static inline void swap_ushort(unsigned short &x)
{
	x = (x << 8) | (x >> 8);
}

static inline void swap_vec(vec &x) {
	swap_float(x[0]);
	swap_float(x[1]);
	swap_float(x[2]);
}


// Figure out whether this machine is little- or big-endian
static bool we_are_little_endian()
{
	// The following appears to be legal according to
	// C99 strict-aliasing rules
	int tmp = 1;
	return !!(* (unsigned char *) &tmp);
}


// Read data from a .xsh file
bool xsh_read(const char *filename, TriMesh *mesh, XSHdata *data)
{
	int nv, nts, nlev;
	bool need_swap = !we_are_little_endian();

	FILE *f = fopen(filename, "rb");
	if (!f) {
		fprintf(stderr, "Couldn't open file %s\n", filename);
		return false;
	}

	// Read in header
	const int xsh_version = 1;
	char hbuf[4];
	if (!fread(hbuf, 4, 1, f)) {
		fprintf(stderr, "Couldn't read file header\n");
		goto err;
	}
	if (hbuf[0] != 'X' || hbuf[1] != 'S' || hbuf[2] != 'H') {
		fprintf(stderr, "Not an XSH file\n");
		goto err;
	}
	if (hbuf[3] != xsh_version) {
		fprintf(stderr, "Wrong XSH version\n");
		goto err;
	}


	// Read in bsphere
	if (!fread(&mesh->bsphere.center, 12, 1, f)) goto err;
	if (!fread(&mesh->bsphere.r, 4, 1, f)) goto err;
	if (need_swap) {
		swap_vec(mesh->bsphere.center);
		swap_float(mesh->bsphere.r);
	}
	mesh->bsphere.valid = true;


	// Read in vertices
	if (!fread(&nv, 4, 1, f)) goto err;
	if (need_swap) swap_int(nv);

	{
	vector<unsigned char> cvertices(9*nv);
	if (!fread(&cvertices[0], cvertices.size(), 1, f)) goto err;
	uncompress_vertices(mesh->vertices, cvertices, mesh->bsphere);
	}


	// Read in tstrips
	if (!fread(&nts, 4, 1, f)) goto err;
	if (need_swap) swap_int(nts);

	{
	vector<unsigned char> ctstrips(nts);
	if (!fread(&ctstrips[0], ctstrips.size(), 1, f)) goto err;
	uncompress_tstrips(mesh->tstrips, ctstrips);
	}
	mesh->convert_strips(TriMesh::TSTRIP_LENGTH);


	// Read in # levels
	if (!fread(&nlev, 4, 1, f)) goto err;
	if (need_swap) swap_int(nlev);

	data->sigmas.resize(nlev);
	data->snormals.resize(nlev);
	data->spdir.resize(nlev);
	data->smeancurv.resize(nlev);


	// For each level...
	for (int i = 0; i < nlev; i++) {
		// Read in sigma
		if (!fread(&data->sigmas[i], 4, 1, f)) goto err;
		if (need_swap) swap_float(data->sigmas[i]);

		// Read in data
		vector<unsigned> cnorm(nv);
		vector<unsigned short> cpdir(nv);
		vector<unsigned short> cmeancurv(nv);
		if (!fread(&cnorm[0], 4*nv, 1, f)) goto err;
		if (!fread(&cpdir[0], 2*nv, 1, f)) goto err;
		if (!fread(&cmeancurv[0], 2*nv, 1, f)) goto err;
		if (need_swap) {
			for (int j = 0; j < nv; j++) {
				swap_unsigned(cnorm[j]);
				swap_ushort(cpdir[j]);
				swap_ushort(cmeancurv[j]);
			}
		}

		// Uncompress it...
		uncompress_normals(data->snormals[i], cnorm);
		uncompress_pdir(data->spdir[i], cpdir, data->snormals[i], cnorm);
		uncompress_meancurv(data->smeancurv[i], cmeancurv);
	}

	// Read in colors if we have them
	{
	vector<unsigned char> colordata(3*nv);
	if (fread(&colordata[0], 3*nv, 1, f)) {
		mesh->colors.resize(nv);
		for (int i = 0; i < nv; i++)
			mesh->colors[i] = Color(&colordata[3*i]);
	}
	}

	fclose(f);
	return true;

err:
	fprintf(stderr, "Error reading file %s\n", filename);
	fclose(f);
	return false;
}


// Write data to a .xsh file
void xsh_write(const char *filename, TriMesh *mesh, XSHdata *data)
{
	bool need_swap = !we_are_little_endian();

	FILE *f = fopen(filename, "wb");

	// Write out header
	const int xsh_version = 1;
	fprintf(f, "XSH%c", xsh_version);

	// Write out bsphere
	if (need_swap) {
		swap_vec(mesh->bsphere.center);
		swap_float(mesh->bsphere.r);
	}
	fwrite(&mesh->bsphere.center, 12, 1, f);
	fwrite(&mesh->bsphere.r, 4, 1, f);
	if (need_swap) {
		swap_vec(mesh->bsphere.center);
		swap_float(mesh->bsphere.r);
	}

	// Write out vertices
	int nv = mesh->vertices.size();
	if (need_swap) swap_int(nv);
	fwrite(&nv, 4, 1, f);
	if (need_swap) swap_int(nv);

	vector<unsigned char> cvertices;
	compress_vertices(mesh->vertices, cvertices, mesh->bsphere);
	fwrite(&cvertices[0], cvertices.size(), 1, f);

	// Write out tstrips
	vector<unsigned char> ctstrips;
	mesh->convert_strips(TriMesh::TSTRIP_TERM);
	compress_tstrips(mesh->tstrips, ctstrips);
	mesh->convert_strips(TriMesh::TSTRIP_LENGTH);
	int nts = ctstrips.size();
	if (need_swap) swap_int(nts);
	fwrite(&nts, 4, 1, f);
	fwrite(&ctstrips[0], ctstrips.size(), 1, f);

	// Write out # levels
	int nlev = data->sigmas.size();
	if (need_swap) swap_int(nlev);
	fwrite(&nlev, 4, 1, f);
	if (need_swap) swap_int(nlev);


	vector<unsigned> cnorm(nv);
	vector<unsigned short> cpdir(nv);
	vector<unsigned short> cmeancurv(nv);

	// For each level...
	for (int i = 0; i < nlev; i++) {
		// Write out sigma
		if (need_swap) swap_float(data->sigmas[i]);
		fwrite(&data->sigmas[i], 4, 1, f);
		if (need_swap) swap_float(data->sigmas[i]);

		// Compress data...
		compress_normals(data->snormals[i], cnorm);
		compress_pdir(data->spdir[i], cpdir, data->snormals[i], cnorm);
		compress_meancurv(data->smeancurv[i], cmeancurv);
		if (need_swap) {
			for (int j = 0; j < nv; j++) {
				swap_unsigned(cnorm[j]);
				swap_ushort(cpdir[j]);
				swap_ushort(cmeancurv[j]);
			}
		}

		// Write it out
		fwrite(&cnorm[0], 4*nv, 1, f);
		fwrite(&cpdir[0], 2*nv, 1, f);
		fwrite(&cmeancurv[0], 2*nv, 1, f);
	}

	// Write out colors if we have them
	if (!mesh->colors.empty()) {
		vector<unsigned char> colordata(3*nv);
		for (int i = 0; i < nv; i++) {
			colordata[3*i  ] =
				clamp(int(255.0f*mesh->colors[i][0]+0.5f),
				0, 255);
			colordata[3*i+1] =
				clamp(int(255.0f*mesh->colors[i][1]+0.5f),
				0, 255);
			colordata[3*i+2] =
				clamp(int(255.0f*mesh->colors[i][2]+0.5f),
				0, 255);
		}
		fwrite(&colordata[0], 3*nv, 1, f);
	}

	fclose(f);
}

